# -*- coding: utf-8 -*-

##############################################################################
# TTYLight-Python: easily stylize terminal text
##############################################################################
#
# Author: Travis Cardwell <travis.cardwell@yuzutechnology.com>
# Copyright: Copyright (c) 2013, Yuzu Technology, Inc.
#
# This file is subject to the terms and conditions defined in the License
# file included with this source code.
#
##############################################################################

"""\
ttylight Command-Line Utility
=============================

The ``ttylight`` command-line utility highlights string or regular expression
targets in streams of text.

See the :doc:`README <../README>` or run ``ttylight --help`` for usage
information.

API
---
"""


import argparse
import re
import sys

import ttylight


COLORS = [
    ('black',   ttylight.Color.BLACK),
    ('red',     ttylight.Color.RED),
    ('green',   ttylight.Color.GREEN),
    ('yellow',  ttylight.Color.YELLOW),
    ('blue',    ttylight.Color.BLUE),
    ('magenta', ttylight.Color.MAGENTA),
    ('cyan',    ttylight.Color.CYAN),
    ('white',   ttylight.Color.WHITE),
    ('default', ttylight.Color.DEFAULT),
]


INTENSITIES = [
    ('normal', ttylight.Intensity.NORMAL),
    ('bold',   ttylight.Intensity.BOLD),
    ('faint',  ttylight.Intensity.FAINT),
]


def color(arg):
    """Parse color arguments

    Colors are specified using case-insensitive prefixes of the following
    strings:

    * 'black'
    * 'blue'
    * 'cyan'
    * 'default'
    * 'green'
    * 'magenta'
    * 'red'
    * 'white'
    * 'yellow'

    :param str arg: string argument from the command-line
    :returns: ttylight.Color constant
    :rtype: int
    """
    arg_lc = arg.lower()
    for name, const in COLORS:
        if name.startswith(arg_lc):
            return const
    raise argparse.ArgumentTypeError("invalid color: {:s}".format(arg))


def intensity(arg):
    """Parse intensity arguments

    Intensities are specified using case-insensitive prefixes of the following
    strings:

    * 'bold'
    * 'faint'
    * 'normal'

    :param str arg: string argument from the command-line
    :returns: ttylight.Intensity constant
    :rtype: int
    """
    arg_lc = arg.lower()
    for name, const in INTENSITIES:
        if name.startswith(arg_lc):
            return const
    raise argparse.ArgumentTypeError("invalid intensity: {:s}".format(arg))


def target(arg):
    """Parse target arguments

    Targets are either regular expressions (in ``s/.../`` format) or strings.

    :param str arg: string argument from the command line
    :returns: regular expression string for the target
    :rtype: str
    """
    if arg.startswith('s/') and arg.endswith('/'):
        arg = arg[2:-1]
    else:
        arg = re.escape(arg)
    return arg


def print_chart():
    """Print the style chart to STDOUT

    This function displays a chart of all possible combinations of foreground
    colors, background colors, and intensities in a chart.
    """
    colors = [(color.upper().center(9), c) for color, c in COLORS]
    print("TTYLight-Python Chart")
    for intensity, i in INTENSITIES:
        print('\n[{:s}]'.format(intensity.upper()))
        for fgstr, fg in colors:
            line = ''
            for _bgstr, bg in colors:
                line += ttylight.stylize(fgstr, fg, bg, i)
            print(line)


def pipe():
    """Pipe all lines from STDIN to STDOUT

    This function is used when no targets are given.
    """
    for line in sys.stdin:
        sys.stdout.write(line)


def highlight(regex, fg, bg, intensity):
    """Print highlighted lines from STDIN to STDOUT

    This function loops through the lines of ``STDIN``, highlights any targets
    that are in the line, and writes the result to ``STDOUT``.

    :param regex regex: compiled regular expression of what to highlight
    :param int fg: foreground color (ttylight.Color constant)
    :param int bg: background color (ttylight.Color constant)
    :param int intensity: intensity (ttylight.Intensity constant)
    """
    for line in sys.stdin:
        even = True
        for part in regex.split(line):
            if even:
                sys.stdout.write(part)
            else:
                sys.stdout.write(ttylight.stylize(part, fg, bg, intensity))
            even = not even


def epilog():
    """Calculate the --help epilog text

    This function renders the lists of colors and intensities for display in
    the help text.

    :returns: --help epilog text
    :rtype: str
    """
    fst = lambda x: x[0]
    return ("Colors: {" + ', '.join(map(fst, COLORS)) +
            "} Intensities: {" + ', '.join(map(fst, INTENSITIES)) + "}")


def main(clargs=None):
    """Parse command-line arguments and dispatch

    :param [str] clargs: list of command-line arguments (default: sys.argv)
    """
    parser = argparse.ArgumentParser(
        description="ttylight Command-Line Utility",
        epilog=epilog())
    parser.add_argument('-v', '--version', action='store_true',
                        help="show version number and exit")
    parser.add_argument('-c', '--chart', action='store_true',
                        help="show color chart and exit")
    parser.add_argument('-f', '--foreground', metavar='COLOR',
                        type=color, default=ttylight.Color.YELLOW,
                        help="foreground color (default: yellow)")
    parser.add_argument('-b', '--background', metavar='COLOR',
                        type=color, default=ttylight.Color.RED,
                        help="background color (default: red)")
    parser.add_argument('-i', '--intensity', metavar='INTENSITY',
                        type=intensity, default=ttylight.Intensity.BOLD,
                        help="intensity (default: bold)")
    parser.add_argument('targets', metavar='TARGET', nargs='*',
                        type=target,
                        help="str|s/regex/")
    args = parser.parse_args(clargs)
    if args.version:
        print('TTYLight-Python {:s}'.format(ttylight.__version__))
    elif args.chart:
        print_chart()
    else:
        try:
            if len(args.targets) < 1:
                pipe()
            else:
                highlight(re.compile('(' + '|'.join(args.targets) + ')'),
                          args.foreground, args.background, args.intensity)
        except KeyboardInterrupt:  # pragma: no cover
            pass


if __name__ == '__main__':  # pragma: no cover
    main()
