# -*- coding: utf-8 -*-

##############################################################################
# TTYLight-Python: easily stylize terminal text
##############################################################################
#
# Author: Travis Cardwell <travis.cardwell@yuzutechnology.com>
# Copyright: Copyright (c) 2013, Yuzu Technology, Inc.
#
# This file is subject to the terms and conditions defined in the License
# file included with this source code.
#
##############################################################################

"""\
ttylight.ttylog Library
=======================

The :mod:`ttylight.ttylog` library is a logging library that stylizes terminal
logs and provides :mod:`helpers <ttylight.ttylog.helpers>` for many common
logging tasks.  The library provides :func:`info`, :func:`warn`,
:func:`error`, and :func:`custom` logging functions, all of which can be
customized using :meth:`State.customize`.

Here is an (contrived) example of using this library without customization::

    import os
    from ttylight import ttylog

    def read_file(fpath, content=''):
        if not os.path.exists(fpath):
            ttylog.warn(fpath + " not found")
        elif not os.path.isfile(fpath):
            ttylog.warn(fpath + " is not a file")
        elif not os.access(fpath, os.R_OK):
            ttylog.warn(fpath + " cannot be read")
        else:
            try:
                content = open(fpath).read()
            except OSError as e:
                ttylog.warn("error reading " + fpath)
                ttylog.info(str(e))
        return content

    if __name__ == '__main__':
        print(read_file('test.txt', "default content\\n"))

Here is an example of how to change the above code to customize the
:func:`warn` function so that it displays "warning:" on a red background::

    import os
    import ttylight
    from ttylight import ttylog

    def read_file(fpath, content=''):
        if not os.path.exists(fpath):
            ttylog.warn(fpath + " not found")
        elif not os.path.isfile(fpath):
            ttylog.warn(fpath + " is not a file")
        elif not os.access(fpath, os.R_OK):
            ttylog.warn(fpath + " cannot be read")
        else:
            try:
                content = open(fpath).read()
            except OSError as e:
                ttylog.warn("error reading " + fpath)
                ttylog.info(str(e))
        return content

    if __name__ == '__main__':
        ttylog.State.customize(ttylog.State.WARN, bg=ttylight.Color.RED)
        print(read_file('test.txt', "default content\\n"))

For more details about customization, see :meth:`State.customize`, and for
examples of prefix/suffix implementations, see the
:mod:`helpers <ttylight.ttylog.helpers>` module.

API
---
"""

import re
import sys

import ttylight


class State:
    """ttylight.ttylog Module State

    This class is used as a namespace and is not meant to be instantiated.  It
    encapsulates the module state in class variables and provides class
    methods for changing the state.

    The class also defines class variable constants for referencing the module
    functions.

    :var int INFO: info function
    :var int WARN: warn function
    :var int ERROR: error function
    :var int CUSTOM: custom function
    :var int ALL: info, warn, error, and custom functions
    """

    _ansi = None

    _custom = {}

    INFO = 0
    WARN = 1
    ERROR = 2
    CUSTOM = 3
    ALL = 999

    @classmethod
    def enable_ansi(cls, ansi=True):
        """Enable/disable ANSI escape sequences

        When a terminal that does not support ANSI escape sequences is
        being used, they should be disabled.  This disables all stylization of
        output text.

        The library automatically enables/disables ANSI when loaded, based on
        the environment.  This function can be called to override the
        automatic settings.

        :param bool ansi: True to enable ANSI, False to disable
        """
        cls._ansi = ansi

    @classmethod
    def _auto_ansi(cls, force=True):
        """Automatically enable/disable ANSI based on the environment

        :param bool force: False to only set cls._ansi if not done before
        """
        if force or cls._ansi is None:
            try:
                import curses
                curses.initscr()
                cls._ansi = curses.has_colors()
                curses.endwin()
            except ImportError:  # pragma: no cover
                # this block is not testable on Linux
                cls._ansi = False

    @classmethod
    def _defaults(cls, func):
        """Get the defaults for a function

        Function settings are stored in a dictionary with the following items:

        * 'enable': True to enable the function, False to disable it
        * 'stream': stream to write to
        * 'fg': prefix foreground color
        * 'bg': prefix background color
        * 'intensity': prefix intensity
        * 'prefix': prefix str|callable
        * 'prefix_space': True to insert a space after the prefix
        * 'suffix': suffix str|callable
        * 'suffix_space': True to insert a space before the suffix
        * 'title': True to set terminal title

        :param int func: log function (State.{INFO|WARN|ERROR|CUSTOM})
        :returns: function settings
        :rtype: dict
        """
        defaults = {
            'enable': True,
            'stream': sys.stderr,
            'bg': ttylight.Color.DEFAULT,
            'intensity': ttylight.Intensity.BOLD,
            'prefix_space': True,
            'suffix': None,
            'suffix_space': True,
            'title': False,
        }
        if func == cls.INFO:
            defaults['fg'] = ttylight.Color.GREEN
            defaults['prefix'] = 'info:'
        elif func == cls.WARN:
            defaults['fg'] = ttylight.Color.YELLOW
            defaults['prefix'] = 'warning:'
        elif func == cls.ERROR:
            defaults['fg'] = ttylight.Color.RED
            defaults['prefix'] = 'error:'
        elif func == cls.CUSTOM:
            defaults['fg'] = ttylight.Color.BLUE
            defaults['prefix'] = None
        else:
            raise Exception("invalid func")
        return defaults

    @classmethod
    def customize(cls, func, reset=False, enable=None, stream=None, fg=None,
                  bg=None, intensity=None, prefix=None, prefix_space=None,
                  suffix=None, suffix_space=None, title=None):
        """Customize a logging function

        Customize the behavior of one or all of the :func:`info`,
        :func:`warn`, :func:`error`, and :func:`custom` functions using a
        :mod:`State` constant (``INFO|WARN|ERROR|CUSTOM|ALL``) to select which
        function(s) to customize.

        The ``reset`` parameter determines the values to use for parameters
        that are not provided.  When ``reset=False``, any previous
        customizations that are not overridden are maintained.  When
        ``reset=True``, function defaults take priority over previous
        customizations.  The following code can be used to reset functions to
        the default settings::

            from ttylight import ttylog

            ttylog.State.customize(ttylog.State.ALL, reset=True)

        Use the ``enable`` parameter to enable or disable a function.  By
        default, all functioned are enabled.  For example, the following code
        disables the :func:`info` function::

            from ttylight import ttylog

            ttylog.State.customize(ttylog.State.INFO, enable=False)

        Use the ``stream`` parameter to specify the output stream to write to.
        For example, the following code makes all of the logging functions
        output to ``STDOUT`` instead of ``STDERR``::

            import sys
            from ttylight import ttylog

            ttylog.State.customize(ttylog.State.ALL, stream=sys.stdout)
            ttylog.warn("test")

        Specify the foreground :class:`Color <ttylight.Color>`, the background
        :class:`Color <ttylight.Color>`, and the
        :class:`Intensity <ttylight.Intensity>` used to stylize the log prefix
        using the ``fg``, ``bg``, and ``intensity`` parameters.  For example,
        the following code makes the :func:`info` function use default colors
        and normal intensity::

            import ttylight
            from ttylight import ttylog

            ttylog.State.customize(ttylog.State.INFO,
                                   fg=ttylight.Color.DEFAULT,
                                   bg=ttylight.Color.DEFAULT,
                                   intensity=ttylight.Intensity.NORMAL)
            ttylog.warn("test")

        Specify a custom prefix as a string, an object with a ``__repr__``
        method, or a function that returns a string.  For example, the
        following code localizes the :func:`warn` function to Spanish::

            from ttylight import ttylog

            ttylog.State.customize(ttylog.State.WARN, prefix="advertencia:")
            ttylog.warn("no hay nada mas")

        This outputs "advertencia: no hay nada mas" with "advertencia:"
        stylized according to the settings for the :func:`warn` function.

        Specify a custom suffix as a string, an object with a ``__repr__``
        method, or a function that returns a string.  For example, the
        following code appends a timestamp to call logging functions::

            from datetime import datetime
            from ttylight import ttylog

            ttylog.State.customize(
                ttylog.State.ALL,
                suffix=lambda: '[' + datetime.now().isoformat() + ']')
            ttylog.error("big problems")

        By default, a space is automatically inserted after the prefix and
        before the suffix.  This can be avoided by setting
        ``prefix_space=False`` or ``suffix_space=True``.

        For more complex prefix/suffix examples, see the
        :mod:`helpers <ttylight.ttylog.helpers>` module.

        By specifying ``title=True``, the output of a logging function is
        also output in the terminal title.  As an example use case, a program
        can use the :func:`custom` logging function to log progress
        information, and the progress information can be displayed in the
        title so that it is visible even when the terminal is minimized.

        :param int func: log function (State.{INFO|WARN|ERROR|CUSTOM|ALL})
        :param bool reset: True to reset any non-specified properties
        :param file stream: stream to output to
        :param int fg: foreground color (ttylight.Color constant)
        :param int bg: background color (ttylight.Color constant)
        :param int intensity: intensity (ttylight.Intensity constant)
        :param str|callable prefix: log prefix
        :param str|callable suffix: log suffix
        :param bool title: True to output to title, False to not
        """
        params = {}
        if enable is not None:
            params['enable'] = enable
        if stream is not None:
            params['stream'] = stream
        if fg is not None:
            params['fg'] = fg
        if bg is not None:
            params['bg'] = bg
        if intensity is not None:
            params['intensity'] = intensity
        if prefix is not None:
            params['prefix'] = prefix
        if prefix_space is not None:
            params['prefix_space'] = prefix_space
        if suffix is not None:
            params['suffix'] = suffix
        if suffix_space is not None:
            params['suffix_space'] = suffix_space
        if title is not None:
            params['title'] = title
        clear = reset and (len(params) == 0)
        if func == cls.ALL:
            funcs = [cls.INFO, cls.WARN, cls.ERROR, cls.CUSTOM]
        else:
            funcs = [func]
        for func in funcs:
            if clear and func in cls._custom:
                del cls._custom[func]
            else:
                if reset or func not in cls._custom:
                    cls._custom[func] = cls._defaults(func)
                cls._custom[func].update(params)

    @classmethod
    def _get(cls, func):
        """Get current function settings

        :param int func: log function (State.{INFO|WARN|ERROR|CUSTOM})
        :returns: function settings
        :rtype: dict
        """
        if func not in cls._custom:
            return cls._defaults(func)
        return cls._custom[func]


def _render_string(string):
    """Render a prefix/suffix as a string

    :param str|callable string: log prefix/suffix
    :returns: prefix/suffix string
    :rtype: str
    """
    if callable(string):
        string = string()
    if string is None:
        return ''
    if not isinstance(string, str):
        string = str(string)
    return string


def _filter_ansi(text):
    """Filter ANSI color sequences from text

    :param str text: text to filter
    :returns: filtered text
    :rtype: str
    """
    re_ansi = re.compile('\033\\[[^a-zA-Z]*?[a-zA-Z]')
    return ''.join([part for part in re_ansi.split(text)])


def _render_output(msg, settings):
    """Render a line of output, writing it to the stream

    :param str msg: message to log
    :param dict settings: function settings
    """
    if settings['enable']:
        out = ttylight.stylize(_render_string(settings['prefix']),
                               settings['fg'], settings['bg'],
                               settings['intensity'])
        if out != '' and settings['prefix_space']:
            out += ' '
        out += msg
        suffix = _render_string(settings['suffix'])
        if suffix != '':
            if settings['suffix_space']:
                out += ' '
            out += suffix
        if State._ansi is not True:
            out = _filter_ansi(out)
        elif settings['title']:
            ttylight.set_terminal_title(_filter_ansi(out))
        settings['stream'].write(out + '\n')


def info(msg):
    """Log an informational message to the terminal

    By default, this functions writes the message to ``STDERR``, prefixed
    with "info:" colored green with a bold intensity.  For example::

        import ttylog

        ttylog.info("this is an informational message")

    The behavior of this function can be customized using
    :meth:`State.customize` using the ``State.INFO`` constant.

    :param str msg: message to log
    """
    _render_output(msg, State._get(State.INFO))


def warn(msg):
    """Log a warning message to the terminal

    By default, this functions writes the message to ``STDERR``, prefixed
    with "warning:" colored yellow with a bold intensity.  For example::

        import ttylog

        ttylog.warn("this is a warning message")

    The behavior of this function can be customized using
    :meth:`State.customize` using the ``State.WARN`` constant.

    :param str msg: message to log
    """
    _render_output(msg, State._get(State.WARN))


def error(msg):
    """Log an error message to the terminal

    By default, this functions writes the message to ``STDERR``, prefixed
    with "error:" colored red with a bold intensity.  For example::

        import ttylog

        ttylog.error("this is an error message")

    The behavior of this function can be customized using
    :meth:`State.customize` using the ``State.ERROR`` constant.

    :param str msg: message to log
    """
    _render_output(msg, State._get(State.ERROR))


def custom(msg, prefix=None, suffix=None):
    """Log a custom message to the terminal

    By default, this functions writes the message to ``STDERR`` with no prefix
    or suffix.  When a prefix is provided, it is colored blue with a bold
    intensity by default.

    Specify a custom prefix or suffix as a string, an object with a
    ``__repr__`` method, or a function that returns a string.  For example::

        from datetime import datetime
        import ttylog

        ttylog.custom("this is a custom message",
                      prefix="status:",
                      suffix=lambda: '[' + datetime.now().isoformat() + ']')

    The behavior of this function can be customized using
    :meth:`State.customize` using the ``State.CUSTOM`` constant.

    :param str msg: message to log
    :param str|callable prefix: log prefix
    :param str|callable suffix: log suffix
    """
    settings = State._get(State.CUSTOM).copy()
    if prefix is not None:
        settings['prefix'] = prefix
    if suffix is not None:
        settings['suffix'] = suffix
    _render_output(msg, settings)


# automatically enable/disable ANSI when the module is loaded
State._auto_ansi(force=False)
