# -*- coding: utf-8 -*-

##############################################################################
# TTYLight-Python: easily stylize terminal text
##############################################################################
#
# Author: Travis Cardwell <travis.cardwell@yuzutechnology.com>
# Copyright: Copyright (c) 2013, Yuzu Technology, Inc.
#
# This file is subject to the terms and conditions defined in the License
# file included with this source code.
#
##############################################################################

"""\
ttylight.ttylog Helpers
=======================

Helpers provide common terminal logging functionality for use in the
:mod:`ttylight.ttylog` logging library.  The following helpers are provided:

* :class:`Labeller`: displays labels
* :class:`Spinner`: displays *spinner*-style animations
* :class:`Counter`: displays a counter with an optional end count
* :class:`Timestamper`: displays timestamps
* :class:`ProgressTracker`: displays (estimated) remaining time, etc.

Create your own log helper by extending the :class:`LogHelper` class.  See
the source code of the above classes for concrete examples.

The :class:`CompositeHelper` class allows you to combine any number of log
helpers, including those that you write yourself.

Here is an example that uses :func:`ttylog.custom <ttylight.ttylog.custom>`
to add a counter prefix and timestamp suffix to custom logs::

    import time
    from ttylight import ttylog
    from ttylight.ttylog.helpers import Counter, Timestamper

    ttylog.State.customize(ttylog.State.CUSTOM, prefix=Counter(width=2),
                           suffix=Timestamper(prefix='(', suffix=')'))
    for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
        ttylog.custom(c)
        time.sleep(0.2)

API
---
"""

import datetime
import math

from ttylight import ttylog


def render_timedelta(delta, fmt=None):
    """Render a timedelta with the given format

    The following formats are supported:

    * 'S.UUU'
    * 'M:SS.UUU'
    * 'M:SS'
    * 'H:MM:SS'
    * 'DdHH:MM:SS'
    * 'DdHH:MM'

    :param datetime.timedelta delta: timedelta to render
    :param str fmt: format (default/invalid: 'H:MM:SS')
    :returns: the rendered timedelta
    :rtype: str
    """
    tsecs = delta.total_seconds()
    if fmt == 'S.UUU':
        return '{:0.03f}'.format(tsecs)
    if fmt == 'M:SS.UUU':
        mins, _secs = divmod(math.floor(tsecs), 60)
        secs = tsecs - 60 * mins
        return '{:d}:{:06.03f}'.format(mins, secs)
    mins, secs = divmod(int(round(tsecs)), 60)
    if fmt == 'M:SS':
        return '{:d}:{:02d}'.format(mins, secs)
    hours, mins = divmod(mins, 60)
    if fmt == 'H:MM:SS':
        return '{:d}:{:02d}:{:02d}'.format(hours, mins, secs)
    days, hours = divmod(hours, 24)
    if fmt == 'DdHH:MM:SS':
        return '{:d}d{:02d}:{:02d}:{:02d}'.format(days, hours, mins, secs)
    if fmt == 'DdHH:MM':
        return '{:d}d{:02d}:{:02d}'.format(days, hours, mins)
    return '{:d}:{:02d}:{:02d}'.format(days * 24 + hours, mins, secs)


class LogHelper:
    """Log helper base class

    Subclass this class to create log helpers.  Be sure to call the
    :class:`LogHelper` constructor from the subclass constructor.  The
    :meth:`render` and :meth:`render_fields` methods must both be implemented
    in the subclass.

    If a subclass has state that can be reset, then override the :meth:`reset`
    method.  Be sure to call ``super().reset()``.

    If a subclass needs to perform operations before or after rendering,
    override the :meth:`render_pre` or :meth:`render_post` methods,
    respectively.

    The ``sameline`` parameter causes the helper to output an ANSI sequence
    that results in multiple lines of the output (printed consecutively) to be
    printed on the same line.  This feature should only be used with log
    prefixes.

    For concrete examples, please see the implementations of the included
    helpers.
    """

    def __init__(self, sameline):
        """Initialize a log helper

        This method initializes the :class:`LogHelper` ``sameline``
        functionality.  Be sure to call this method from any subclass
        constructors.

        :param bool sameline: True to display on single line
        """
        self._sameline = sameline and ttylog.State._ansi
        self._sameline_first = True

    def reset(self):
        """Reset a log helper

        This method resets ``sameline``.  Be sure to call this method from any
        subclass ``reset`` methods.
        """
        self._sameline_first = True

    def render_pre(self):
        """Perform operations before rendering

        Override this method in subclasses to perform operations before
        rendering.  By default, it does nothing.
        """
        pass

    def render(self):
        """Render the helper

        Override this method in subclasses to define how to render them.  By
        default, an empty string is returned.

        :returns: string representation of the helper
        :rtype: str
        """
        return ''

    def render_fields(self):
        """Render all helper fields separately

        Override this method in subclasses to define how to render each field.
        By default, not fields are rendered.

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        return {}

    def render_post(self):
        """Perform operations after rendering

        Override this method in subclasses to perform operations after
        rendering.  By default, it does nothing.
        """
        pass

    def __repr__(self):
        """Render the helper for output

        This method runs :meth:`render_pre`, renders the helper with
        :meth:`render`, runs :meth:`render_post`, and does ``sameline``
        processing.  It does not need to be overridden in normal cases.

        :returns: output representation of the helper
        :rtype: str
        """
        self.render_pre()
        s = self.render()
        self.render_post()
        if self._sameline:
            if self._sameline_first:
                self._sameline_first = False
            else:
                s = '\033[F\033[K' + s
        return s


class AttrDict(dict):
    """Helper class to allow attribute access to dictionaries
    """

    def __init__(self, *args, **kwargs):
        """Initialize by pointing __dict__ to self
        """
        super().__init__(*args, **kwargs)
        self.__dict__ = self


class CompositeHelper(LogHelper):
    """Log helper that composites other log helpers

    A :class:`CompositeHelper` log helper allows composition of multiple log
    helpers.

    The ``template`` parameter specifies how to format the output string when
    rendering.  It is formatted using `standard string formatting
    <http://docs.python.org/3.3/library/string.html#string-formatting>`_.
    Composited log helper fields are referenced using attribute syntax.  For
    example, the ``count`` field of a :class:`Counter` log helper registered
    with log helper ID ``c`` can be referenced as ``{c.count}``.

    Composited log helpers must be registered using the :meth:`register`
    method.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import (CompositeHelper, Spinner,
                                             Counter, Timestamper)

        helper = CompositeHelper(
            '{s.state} {t.dt_current} [{c.count}/{c.end}]', sameline=True)
        helper.register('s', Spinner(sameline=False))
        helper.register('t', Timestamper('%H:%M:%S'))
        helper.register('c', Counter(end=26))
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=helper)
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            ttylog.custom(c)
            time.sleep(0.2)
    """

    def __init__(self, template, sameline=False):
        """Initialize a composite helper

        :param str template: render template
        :param bool sameline: True to display on single line (default: False)
        """
        super().__init__(sameline)
        self._template = template
        self._helpers = {}

    def reset(self):
        """Reset the composite

        Calling this methods delegates to all composited log helpers.
        """
        super().reset()
        for lh in self._helpers.values():
            lh.reset()

    def render_pre(self):
        """Perform operations before rendering

        Calling this methods delegates to all composited log helpers.
        """
        for lh in self._helpers.values():
            lh.render_pre()

    def render(self):
        """Render the composite

        :returns: rendered composite at the current state
        :rtype: str
        """
        fields = {}
        for lhid, lh in self._helpers.items():
            fields[lhid] = AttrDict(lh.render_fields())
        return self._template.format(**fields)

    def render_fields(self):
        """Render the composite fields

        This class provides the following fields:

        * composite: the rendered composite

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        return {'composite': self.render()}

    def render_post(self):
        """Perform operations after rendering

        Calling this methods delegates to all composited log helpers.
        """
        for lh in self._helpers.values():
            lh.render_post()

    def register(self, lhid, lh):
        """Register a log helper

        :param str lhid: log helper ID for use in the template
        :param LogHelper lh: log helper
        """
        self._helpers[lhid] = lh


class Labeller(LogHelper):
    """Log helper that displays labels

    A :class:`Labeller` log helper displays labels in logs.

    The initial label is set using the ``label`` parameter of the constructor.
    The label can then be changed at any time using the :meth:`set_label`
    method.

    The ``align`` parameter sets the alignment of the labels within a field
    width specified by the ``width`` parameter.  ``align`` should be one of
    the alignment class variable constants defined in this class.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the label field.  Empty strings are
    used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        from ttylight import ttylog
        from ttylight.ttylog.helpers import Labeller

        labeller = Labeller('low', align=Labeller.RIGHT, width=6, suffix=':')
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=labeller)
        ttylog.custom('1')
        ttylog.custom('2')
        labeller.set_label('mid')
        ttylog.custom('3')
        ttylog.custom('4')
        labeller.set_label('high')
        ttylog.custom('5')
        ttylog.custom('6')

    :var int LEFT: left alignment constant
    :var int CENTER: center alignment constant
    :var int RIGHT: right alignment constant
    """

    LEFT = 0
    CENTER = 1
    RIGHT = 2

    def __init__(self, label, align=0, width=None, prefix='', suffix='',
                 sameline=False):
        """Initialize a labeller

        :param str label: initial label
        :param int align: {LEFT|CENTER|RIGHT} (default: LEFT)
        :param int width: label field width (default: <auto>)
        :param str prefix: prefix string (default: '')
        :param str suffix: suffix string (default: '')
        :param bool sameline: True to display on single line (default: False)
        """
        super().__init__(sameline)
        self._label = label
        self._align = align
        self._width = width if width is not None else 0
        self._prefix = prefix
        self._suffix = suffix

    def render(self):
        """Render the label

        :returns: rendered label
        :rtype: str
        """
        s = self._prefix
        if self._align == Labeller.RIGHT:
            s += self._label.rjust(self._width)
        elif self._align == Labeller.CENTER:
            s += self._label.center(self._width)
        else:
            s += self._label.ljust(self._width)
        return s + self._suffix

    def render_fields(self):
        """Render the label fields

        This class provides the following fields:

        * label: the rendered label

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        return {'label': self.render()}

    def set_label(self, label):
        """Set the label

        :param str label: new label
        """
        self._label = label


class Spinner(LogHelper):
    """Log helper that displays a spinner

    A :class:`Spinner` log helper loops through a list of states, displaying
    one at a time.  When combined with ``sameline``, it results in a
    *spinner*-style animation.  This helper is generally used as a prefix.  It
    can be used as a suffix if logs are of fixed width, ``sameline=True`` for
    the log prefix, and ``sameline=False`` for the log suffix.

    The ``states`` parameter specifies the states that should be looped
    through when displaying.  If each state is one character long, then the
    parameter can be specified as a string.  If a state is more than one
    character long, then the parameter must be specified as a list of strings.
    A clockwise spinner is displayed by default.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the state field.  Empty strings are
    used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.  It is on
    by default, in order to make an animation.

    Static methods :meth:`Spinner.spinner_states`,
    :meth:`Spinner.pulser_states`, :meth:`Spinner.pacer_states`, and
    :meth:`Spinner.pinball_states` provide various animations to use.

    Static method :meth:`Spinner.widen` can be used to widen animations with
    single-character states to wider states.  It can be used with
    :meth:`Spinner.spinner_states` and :meth:`Spinner.pulser_states`, for
    example.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import Spinner

        ttylog.State.customize(ttylog.State.CUSTOM,
                               prefix=Spinner(Spinner.spinner_states()))
        for i in range(1, 33):
            ttylog.custom(str(i))
            time.sleep(0.2)

    """

    def __init__(self, states=None, prefix='', suffix='', sameline=True):
        """Initialize a spinner

        :param [str] states: spinner states (default: [''])
        :param str prefix: prefix string (default: '')
        :param str suffix: suffix string (default: '')
        :param bool sameline: True to display on single line (default: True)
        """
        super().__init__(sameline)
        if states is None:
            self._states = Spinner.spinner_states()
        else:
            self._states = states
        self._states_len = len(self._states)
        self._states_i = 0
        self._prefix = prefix
        self._suffix = suffix

    def reset(self):
        """Reset the spinner
        """
        super().reset()
        self._states_i = 0

    def render(self):
        """Render the spinner

        :returns: rendered spinner at the current state
        :rtype: str
        """
        s = self._states[self._states_i]
        self._states_i += 1
        if self._states_i >= self._states_len:
            self._states_i = 0
        return self._prefix + s + self._suffix

    def render_fields(self):
        """Render the spinner fields

        This class provides the following fields:

        * state: the rendered spinner at the current state

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        return {'state': self.render()}

    @staticmethod
    def widen(frames, width, right=False):
        """Widen single-character states into wider states

        Example::

            lh = Spinner(Spinner.widen(Spinner.spinner_states(), 4))

        :param str frames: string of single-character states
        :param int width: desired state width
        :param bool right: True to change right faster (default: False)
        :returns: widened statas
        :rtype: [str]
        """
        states = [c for c in frames]
        for i in range(1, width):
            ss = []
            if right:
                for c in frames:
                    ss += [c + s for s in states]
            else:
                for c in frames:
                    ss += [s + c for s in states]
            states = ss
        return states

    @staticmethod
    def spinner_states(cc=False):
        """Get spinner-style animation states

        Example::

            lh = Spinner(Spinner.spinner_states())

        :param bool cc: True to rotate counter-clockwise (default: False)
        :returns: spinner states
        :rtype: [str]
        """
        return '|\\-/' if cc else '|/-\\'

    @staticmethod
    def pulser_states(flat=False):
        """Get pulser-style animation states

        Example::

            lh = Spinner(Spinner.pulser_states())

        :param bool flat: True for a flat pulser (default: False)
        :returns: pulser states
        :rtype: [str]
        """
        return '_-^-' if flat else '.oOo'

    @staticmethod
    def pacer_states(width=5, lside='|', rside='|', left='<', right='>'):
        """Get pacer-style animation states

        Example::

            lh = Spinner(Spinner.pacer_states())

        :param int width: desired state width (default: 5)
        :param str lside: left side character (default: '|')
        :param str rside: right side character (default: '|')
        :param str left: left pace character (default: '<')
        :param str right: right pace character (default: '>')
        :returns: pacer states
        :rtype: [str]
        """
        if width <= 1:
            return lside + right + rside + left
        elif width == 2:
            return [lside + ' ', right + ' ', ' ' + rside, ' ' + left]
        w1 = width - 1
        rstates = []
        lstates = []
        for i in range(1, w1):
            lsp = ' ' * i
            rsp = ' ' * (w1 - i)
            rstates.append(lsp + right + rsp)
            lstates.append(rsp + left + lsp)
        sp = ' ' * w1
        return [lside + sp] + rstates + [sp + rside] + lstates

    @staticmethod
    def pinball_states(width=5, ball='*'):
        """Get pinball-style animation states

        Example::

            lh = Spinner(Spinner.pinball_states())

        :param int width: desired state width (default: 5)
        :param str ball: ball character (default: '*')
        :returns: pinball states
        :rtype: [str]
        """
        if width <= 1:
            return '|\\'
        elif width == 2:
            return ['||', '\\/']
        elif width == 3:
            return ['\\*|', '|*/']
        w3 = width - 3
        rstates = []
        lstates = []
        for i in range(1, w3):
            lsp = ' ' * i
            rsp = ' ' * (w3 - i)
            rstates.append('|' + lsp + ball + rsp + '|')
            lstates.append('|' + rsp + ball + lsp + '|')
        sp = ' ' * w3
        return (['\\' + ball + sp + '|'] + rstates + ['|' + sp + ball + '/'] +
                lstates)


class Counter(LogHelper):
    """Log helper that displays a counter

    A :class:`Counter` log helper implements a counter with an optional end
    count.

    The ``template`` parameter specifies how to format the output string when
    rendering.  It is formatted using `standard string formatting
    <http://docs.python.org/3.3/library/string.html#string-formatting>`_.
    The following fields are available:

    * start: the start count
    * count: the current count
    * end: the end count

    Note that each field in the template is a string; they are formatted as
    strings before being passed to the template.  Use the ``width`` parameter
    to specify a zero-padded width for these values.

    When not specified, and ``end`` is not specified, ``template`` defaults to
    ``[{count}]``.  When not specified, and ``end`` is specified, it defaults
    to ``[{count}/{end}]``.

    The ``start`` parameter can be used to set the starting count.  It is set
    to 0 by default.

    The ``end`` parameter can be used to set the end count.

    The ``width`` parameter can be used to set the width of the counter,
    padding it with zeros.

    The ``inc`` parameter specifies how the counter will be incremented.  By
    default, it is ``Counter.RENDER_PRE``, which causes the counter to be
    incremented before each time it is rendered.  This works well with the
    default ``start`` value of 0: the counter represents the number of items
    counted.  When ``inc`` is set to ``Counter.RENDER_POST``, the counter is
    incremented after each time it is rendered.  This works well with
    ``start`` set to 1.  When ``inc`` is set to ``Counter.MANUAL``, the
    counter is only incremented by calling the :meth:`inc` method.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        from ttylight import ttylog
        from ttylight.ttylog.helpers import Counter

        counter = Counter(end=26)
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=counter)
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            ttylog.custom(c)

    :var int MANUAL: manual increment constant
    :var int RENDER_PRE: increment before rendering constant
    :var int RENDER_POST: increment after rendering constant
    """

    RENDER_PRE = 0
    RENDER_POST = 1
    MANUAL = 2

    def __init__(self, template=None, start=0, end=None, width=None, inc=0,
                 sameline=False):
        """Initialize a counter

        :param str template: render template
        :param int start: starting count, >= 0 (default: 0)
        :param int end: ending count, >= start (default: 1)
        :param int width: width of counter, >= 0 (default: <auto>)
        :param int inc: {RENDER_PRE|RENDER_POST|MANUAL} (default: RENDER_PRE)
        :param bool sameline: True to display on single line (default: False)
        """
        super().__init__(sameline)
        if template is None:
            if end is None:
                self._template = '[{count}]'
            else:
                self._template = '[{count}/{end}]'
        else:
            self._template = template
        self._start = max(start, 0)
        self._count = self._start
        if end is None:
            self._end = max(self._start, 1)
        else:
            self._end = max(end, self._start, 1)
        if width is None:
            self._width = math.floor(math.log10(self._end)) + 1
        else:
            self._width = max(width, 0)
        self._inc = inc

    def reset(self):
        """Reset the counter
        """
        super().reset()
        self._count = self._start

    def render_pre(self):
        """Increment the counter when necessary
        """
        if self._inc == Counter.RENDER_PRE:
            self._count += 1

    def render(self):
        """Render the counter

        :returns: rendered counter at the current state
        :rtype: str
        """
        fields = self.render_fields()
        return self._template.format(**fields)

    def render_fields(self):
        """Render the counter fields

        This class provides the following fields:

        * start: the start count
        * count: the current count
        * end: the end count

        :returns: dictionary of field name (str) to rendered value (str)
                :rtype: dict
        """
        return {
            'start': '{:0{w}d}'.format(self._start, w=self._width),
            'count': '{:0{w}d}'.format(self._count, w=self._width),
            'end': '{:0{w}d}'.format(self._end, w=self._width),
        }

    def render_post(self):
        """Increment the counter when necessary
        """
        if self._inc == Counter.RENDER_POST:
            self._count += 1

    def inc(self, count=1):
        """Increment the counter

        :param int count: number to increment by (default: 1)
        """
        self._count += count


class Timestamper(LogHelper):
    """Log helper that displays timestamps

    A :class:`Timestamper` log helper displays timestamps in logs.

    The ``fmt_dt`` parameter is used to specify the format in which the
    date/time is rendered, using `datetime.datetime.strftime
    <http://docs.python.org/3.3/library/datetime.html
    #strftime-and-strptime-behavior>`_.  The default format is ISO 8601.

    The ``tz`` parameter is used to specify a `time zone
    <http://docs.python.org/3.3/library/datetime.html#tzinfo-objects>`_.
    Local time is used by default.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the label field.  Empty strings are
    used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import Timestamper

        ttylog.State.customize(ttylog.State.CUSTOM, prefix=Timestamper())
        for i in range(1, 33):
            ttylog.custom(str(i))
            time.sleep(0.2)
    """

    def __init__(self, fmt_dt='%Y-%m-%dT%H:%M:%S%z', tz=None, prefix='',
                 suffix='', sameline=False):
        """Initialize a timestamper

        :param str fmt_dt: current time format (default: ISO 8601)
        :param datetime.tzinfo tz: time zone (default: local)
        :param str prefix: prefix string (default: '')
        :param str suffix: suffix string (default: '')
        :param bool sameline: True to display on single line (default: False)
        """
        super().__init__(sameline)
        self._fmt_dt = fmt_dt
        self._tz = tz
        self._prefix = prefix
        self._suffix = suffix

    def render(self):
        """Render the timestamp

        :returns: rendered timestamp
        :rtype: str
        """
        dt_cur = datetime.datetime.now(self._tz)
        return self._prefix + dt_cur.strftime(self._fmt_dt) + self._suffix

    def render_fields(self):
        """Render the timestamper fields

        This class provides the following fields:

        * dt_current: the timestamp for the current time

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        return {'dt_current': self.render()}


class ProgressTracker(Counter):
    """Log helper that tracks progress

    A :class:`ProgressTracker` log helper extends :class:`Counter` to add
    progress tracking.

    The ``template`` parameter specifies how to format the output string when
    rendering.  It is formatted using `standard string formatting
    <http://docs.python.org/3.3/library/string.html#string-formatting>`_.
    The following fields are available:

    * start: the start count
    * count: the current count
    * end: the end count
    * elapse: the elapsed time
    * remain: the remaining time (estimated)
    * dt_start: the start time
    * dt_current: the current time
    * dt_end: the end time (estimated)

    Note that each field in the template is a string; they are formatted as
    strings before being passed to the template.  Use the ``width`` parameter
    to specify a zero-padded width for ``start``, ``count``, and ``end``.

    The default ``template`` is ``{remain}``.

    The ``start`` parameter can be used to set the starting count.  It is set
    to 0 by default.

    The ``end`` parameter can be used to set the end count.  It is set to 100
    by default.

    The ``width`` parameter can be used to set the width of the counter,
    padding it with zeros.

    The ``inc`` parameter specifies how the counter will be incremented.  By
    default, it is ``Counter.RENDER_PRE``, which causes the counter to be
    incremented before each time it is rendered.  This works well with the
    default ``start`` value of 0: the counter represents the number of items
    counted.  When ``inc`` is set to ``Counter.RENDER_POST``, the counter is
    incremented after each time it is rendered.  This works well with
    ``start`` set to 1.  When ``inc`` is set to ``Counter.MANUAL``, the
    counter is only incremented by calling the :meth:`inc` method.

    For an accurate ``remain`` and ``dt_end`` estimates, the counter must be
    maintained, but it is not necessary to use ``count`` in the template.

    The ``fmt_dt`` parameter is used to specify the format in which
    ``dt_start``, ``dt_current``, and ``dt_end`` are rendered, using
    `datetime.datetime.strftime
    <http://docs.python.org/3.3/library/datetime.html
    #strftime-and-strptime-behavior>`_.  The default format is ISO 8601.

    The ``fmt_delta`` parameter is used to specify the format in which
    ``elapse`` and ``remain`` are rendered.  The valid formats are defined by
    the :func:`render_timedelta` function.  The default format is 'H:MM:SS'.

    The ``tz`` parameter is used to specify a `time zone
    <http://docs.python.org/3.3/library/datetime.html#tzinfo-objects>`_.
    Local time is used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import ProgressTracker

        ttylog.State.customize(ttylog.State.CUSTOM,
                               prefix=ProgressTracker(end=26))
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            ttylog.custom(c)
            time.sleep(0.2)
    """

    def __init__(self, template='{remain}', start=0, end=100, width=None,
                 inc=0, fmt_dt='%Y-%m-%dT%H:%M:%S%z', fmt_delta='H:MM:SS',
                 tz=None, sameline=False):
        """Initialize a progress tracker

        :param str template: render template
        :param int start: starting count, >= 0 (default: 0)
        :param int end: ending count, >= start (default: 100)
        :param int width: width of counter, >= 0 (default: <auto>)
        :param int inc: {RENDER_PRE|RENDER_POST|MANUAL} (default: RENDER_PRE)
        :param str fmt_dt: time format (default: ISO 8601)
        :param str fmt_delta: delta format (default: 'H:MM:SS')
        :param datetime.tzinfo tz: time zone (default: local)
        :param bool sameline: True to display on single line (default: False)
        """
        super().__init__(template, start, end, width, inc, sameline)
        self._fmt_dt = fmt_dt
        self._fmt_delta = fmt_delta
        self._tz = tz
        self._dt_start = datetime.datetime.now(self._tz)

    def reset(self):
        """Reset the progress tracker
        """
        super().reset()
        self._dt_start = datetime.datetime.now(self._tz)

    def render_fields(self):
        """Render the progress tracker fields

        This class provides the following fields:

        * start: the start count
        * count: the current count
        * end: the end count
        * elapse: the elapsed time
        * remain: the remaining time (estimated)
        * dt_start: the start time
        * dt_current: the current time
        * dt_end: the end time (estimated)

        :returns: dictionary of field name (str) to rendered value (str)
        :rtype: dict
        """
        dt_cur = datetime.datetime.now(self._tz)
        elapse = dt_cur - self._dt_start
        if self._count > 0:
            remain = elapse.total_seconds()
            remain = datetime.timedelta(
                seconds=remain * self._end / self._count - remain)
        else:
            remain = datetime.timedelta(seconds=0)
        dt_end = dt_cur + remain
        return {
            'start': '{:0{w}d}'.format(self._start, w=self._width),
            'count': '{:0{w}d}'.format(self._count, w=self._width),
            'end': '{:0{w}d}'.format(self._end, w=self._width),
            'elapse': render_timedelta(elapse, self._fmt_delta),
            'remain': render_timedelta(remain, self._fmt_delta),
            'dt_start': self._dt_start.strftime(self._fmt_dt),
            'dt_current': dt_cur.strftime(self._fmt_dt),
            'dt_end': dt_end.strftime(self._fmt_dt),
        }
